#****************************************************************************
# menu.pl -- Perl Menu Support Facility 
#
# Version: 2.1
#
# Author:  Steven L. Kunz
#          Networking & Communications
#          Iowa State University Computation Center
#          Ames, IA  50011
#
# Bugs:    skunz@iastate.edu
# Cheers:  skunz@iastate.edu
#
# Date:    Version 1.0 -- May, 1992 -- Original version 
#          Version 1.1 -- Aug, 1992 -- Minor enhancements, bugfixes
#          Version 1.2 -- Nov, 1992 -- Selection bugfix
#          Version 1.3 -- Dec, 1992 -- "top" and "latch" functions added
#          Version 1.4 -- Apr, 1993 -- "r=refresh" added to bottom line
#          Version 2.0 -- Sep, 1993 -- Radio-button, Multiple-selection,
# 				       shell-escape, new "hot-keys", and
#				       "menu_getstr" routine.
#          Version 2.1 -- Oct, 1993 -- Bug fixes
#
# Notes:   This package requires curseperl
#          (distributed with perl 4.36 in the usub directory).
#
#          Use:
#             &menu_init(1,"title");
#             &menu_item("Topic 1","got_1");
#             &menu_item("Topic 2","got_2");
#             ...
#             &menu_item("Topic n","got_n");
#             $sel_text = &menu_display("Select using arrow keys");
#
# Copyright (C) 1992-93  Iowa State University Computation Center
#                        Ames, Iowa  (USA)
#
# menu.pl is free software; you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation; either version 1, or (at your option) any later version.  
#
# menu.pl is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.  
#
# You should have received a copy of the GNU General Public License along with 
# menu.pl.  If not, write to:
#
#  The Free Software Foundation 
#  675 Mass Ave
#  Cambridge, MA 02139, USA
#
#****************************************************************************

package perlmenu;

$did_initterm = 0;
$curses_application = 0;
$menu_exit_routine = "main'clear";
$menu_shell_text = "";
$menu_is_first_one = 1;
$menu_is_top_one = 0;
$menu_top_activated = 0;
$finding_top = 0;
@selected_action = ();
$window = 0;
$xrow = $xcol = 0;

#**********
#  MENU_CURSES_APPLICATION
#
#  Function:	Indicate application is using curses calls.  If called, 
#		the menu routines will not do initscr and endwin calls 
#		(the application must do them).
#
#  Call format:	&menu_curses_application;
#
#  Arguments:	None
#
#  Returns:	Nothing
#**********
sub main'menu_curses_application { $curses_application = 1; }

#**********
#  MENU_SHELL_COMMAND
#
#  Function:	Enable "!" as shell escape from a menu and indicate the 
#		command that should be issued.
#
#  Call format:	&menu_shell_command("shell-path command");
#
#  Arguments:	Shell path (such as "/bin/csh") to start shell spawn-off.
#		If null value is supplied, shell escape is disabled.
#
#  Returns:	Nothing
#**********
sub main'menu_shell_command {
  ($menu_shell_text) = @_;
}

#**********
#  MENU_QUIT_ROUTINE
#
#  Function:	Specify a "cleanup" routine to be called before a "quit"
#		from the application is processed.
#
#  Call format:	&menu_quit_routine("string");
#
#  Arguments:	String containing name of exit routine to call.
#
#  Returns:	Nothing.
#
#**********
sub main'menu_quit_routine { $menu_exit_routine = "main'@_"; }

#**********
#  MENU_INIT
#
#  Function:	Initialize menu type (numbered or unnumbered), arrays, 
#		title, and "top" flags.
#
#  Call format:	&menu_init([0|1],"Top Title",[0|1]);
#
#  Arguments:   Boolean flag indicating whether or not a arrows and numbers
#               are desired (0=no, 1=yes) and title text (for the top line
#               of menu) and an optional boolean top-menu indicator.
#
#  Returns:	Nothing
#
#  Notes:	1) If the title string begins with a "-" the title is not
#		   presented in reverse-video ("standout") representation.
#               2) If this is the FIRST menu_init call and the optional
#                  third opernd is "1", it is the "top" menu.
#**********
sub main'menu_init {
  ($menu_numbered,$menu_top_title,$menu_is_top_one) = @_;

# Perform initscr if not a curses application
  if (!$curses_application) { $window = &main'initscr; }

# Load "magic sequence" array based on terminal type
  if (!$did_initterm) {		# Get terminal info (if we don't have it).
    &defbell unless defined &bell;

    $ku = &main'getcap('ku');	# Cursor-up
    $kd = &main'getcap('kd');	# Cursor-down
    $kr = &main'getcap('kr');	# Cursor-right
    $kl = &main'getcap('kl');	# Cursor-left
    $cr = &main'getcap('cr');	# Carriage-return
    $nl = &main'getcap('nl');	# New-line
    $ansi_ku = "\033[A";	# Ansi cursor-up (for DEC xterm)
    $ansi_kd = "\033[B";	# Ansi cursor-down (for DEC xterm)
    $ansi_kr = "\033[C";	# Ansi cursor-right (for DEC xterm)
    $ansi_kl = "\033[D";	# Ansi cursor-left (for DEC xterm)

    @magic_seq = ($ku,$ansi_ku,$kd,$ansi_kd,$kl,$ansi_kl,$kr,$ansi_kr,
		  $cr,$nl,"\n",
		  "n","N"," ","p","P","b","B","e","E",
		  "a","A","c","C","m","M","=","/","h","H","?","!");
    $did_initterm = 1;
  }

# Check for title format character.
  $menu_top_title_attr = 0;
  if (substr($menu_top_title,0,1) eq '-') {
    $menu_top_title = substr($menu_top_title,1);
    $menu_top_title_attr = 1;
  }

# Center top title
  if (length($menu_top_title) > $main'COLS) {
    $menu_top_title = substr($menu_top_title,0,$main'COLS);
    $menu_top_title_col = 0;
  }
  else {
    $menu_top_title_col = int($main'COLS/2) - int(length($menu_top_title)/2);
  }

# Enable "top menu" functions if first menu is a top menu.
  if (($menu_is_first_one) && ($menu_is_top_one)) { $menu_top_activated = 1; }
  $menu_is_first_one = 0;

# Init selection array
  @menu_sel_text = ();		# Clear menu arrays
  @menu_sel_action = ();
  @selected_action = ();
  $menu_index = 0;		# Reset flags

  $first_line = 2;
  $last_line = $main'LINES - 3;
  $items_per_screen = $last_line - $first_line + 1;
  $arrow_line = $first_line;
  $menu_top_item = 0;

# Return window value from "initscr" call.
  $window;
}

#***********
#  MENU_ITEM
#
#  Function:	Add an item to the active menu.
#
#  Call format:	&menu_item("What you see","test_rtn");
#
#  Arguments:	String presented in menu, string returned if selected
#
#  Returns:	Number of items currently in the menu.
#***********
sub main'menu_item {
  local($item_text,$item_sel) = @_;

# Load into arrays and adjust index
  @menu_sel_text[$menu_index] = $item_text;
  @menu_sel_action[$menu_index] = $item_sel;
  $menu_index++;
}

#**********
#  MENU_DISPLAY 
#
#  Function:	Display items in menu_sel_text array, allow selection, and
#		return appropriate selection-string.
#
#  Call format:	$sel = &menu_display("Prompt text",$arrow_line,$top_item);
#
#  Arguments:   - Prompt text (for the bottom line of menu).
#		- Line number offset for the arrow (defaults to zero).
#		- Index of top item on screen (defaults to zero).
#
#  Returns:     Selected action string (from second param on &menu_item)
#		%UP%    -- "u"|"U" pressed (or "t"|"T" and looking for top)
#               %EMPTY% -- Nothing in menu to display
#
#  Notes:	1) This routine ALWAYS sets "nocbreak" and "echo" terminal 
#		   modes before returning.
#		2) This routine exits directly (after calling the optional 
#		   "quit" routine) if "q"|"Q" is pressed.
#**********
sub main'menu_display {
  ($menu_prompt,$arrow_spec,$menu_top_item) = @_;
  local($ret);

# Check for no "menu_item" calls.
  $total_items = $#menu_sel_text + 1;
  if ($total_items <= 0) {
    &main'nocbreak;		# ALWAYS turn off "cbreak" mode
    &main'echo;			# ALWAYS turn on "echo"
    return("%EMPTY%");
  }

  &main'clear; $xrow = $xcol = 0;
  $ret=&menu_display_internal(0,$menu_prompt,$arrow_spec,$menu_top_item);
  if ($#_ > 0) { @_[1] = $arrow_spec; }
  if ($#_ > 0) { @_[2] = $menu_top_item; }
  &menu_return_prep;
  $ret;
}

#**********
#  MENU_DISPLAY_RADIO 
#
#  Function:	Display items in "radio-button" style.
#
#  Call format:	$sel = &menu_display_radio("Prompt text","current","Done");
#
#  Arguments:   - Prompt text (for the bottom line of menu).
#               - Current setting (one of the "action-texts" that is the
#                 current setting.
#               - Text for "done with setting" top menu item 
#                 Optional - defaults to "(Accept this setting)"
#
#  Returns:     Selected action string (from second param on &menu_item)
#		%UP%    -- "u"|"U" pressed (or "t"|"T" and looking for top) 
#               %EMPTY% -- Nothing in menu to display
#
#  Notes:	1) This routine ALWAYS sets "nocbreak" and "echo" terminal 
#		   modes before returning.
#		2) This routine exits directly (after calling the optional 
#		   "quit" routine) if "q"|"Q" is pressed.
#**********
sub main'menu_display_radio {
  ($menu_prompt,$current,$done_text) = @_;
  ($last_arrow,$last_top) = 0;

# Check for no "menu_item" calls.
  $total_items = $#menu_sel_text + 1;
  if ($total_items <= 0) {
    &main'nocbreak;		# ALWAYS turn off "cbreak" mode
    &main'echo;			# ALWAYS turn on "echo"
    return("%EMPTY%");
  }

  &main'clear; $xrow = $xcol = 0;

# Insert our top selection item.
  if ($done_text eq "") { $done_text = "(Accept this setting)"; }
  unshift(@menu_sel_text,$done_text);
  unshift(@menu_sel_action,"%DONE%");
  $menu_index++;

  while (1) {
    $ret = &menu_display_internal(1,$menu_prompt,$last_arrow,$last_top,$current);
    last if $ret eq "%DONE%";
    if ($ret eq "%UP%") {
      &menu_return_prep;
      return($ret);
    }
    $current = $ret;
  }
  &menu_return_prep;
  $current;
}

#**********
#  MENU_DISPLAY_MULT
#
#  Function:	Display items in "multiple selection" format.
#
#  Call format:	$sel = &menu_display_mult("Prompt text","Done");
#
#  Arguments:   - Prompt text (for the bottom line of menu)
#               - Text for "done with selections" top menu item 
#                 Optional - defaults to "(Done with selections)"
#
#  Returns:     Selected action string (from second param on &menu_item)
#		%UP%    -- "u"|"U" pressed (or "t"|"T" and looking for top) 
#               %NONE%  -- No items selected
#               %EMPTY% -- Nothing in menu to display
#
#  Notes:	1) This routine ALWAYS sets "nocbreak" and "echo" terminal 
#		   modes before returning.
#		2) This routine exits directly (after calling the optional 
#		   "quit" routine) if "q"|"Q" is pressed.
#**********
sub main'menu_display_mult {
  ($menu_prompt,$done_text) = @_;
  @selected_action = ();
  local($last_arrow,$last_top,$i) = 0;
  local($ret);

# Check for no "menu_item" calls.
  $total_items = $#menu_sel_text + 1;
  if ($total_items <= 0) {
    &main'nocbreak;		# ALWAYS turn off "cbreak" mode
    &main'echo;			# ALWAYS turn on "echo"
    return("%EMPTY%");
  }

  &main'clear; $xrow = $xcol = 0;

# Insert our top selection item.
  if ($done_text eq "") { $done_text = "(Done with selections)"; }
  unshift(@menu_sel_text,$done_text);
  unshift(@menu_sel_action,"%DONE%");
  $menu_index++;

# Loop, allowing toggle of selections, until done.
  while (1) {
    $ret = &menu_display_internal(2,$menu_prompt,$last_arrow,$last_top,0);
    last if $ret eq "%DONE%";
    if (($ret eq "%UP%") || ($ret eq "%EMPTY%")) {
      @selected_action = ();
      &menu_return_prep;
      return($ret);
    }
    if ($selected_action[$item]) { $selected_action[$item] = 0; } # Toggle off
    else { $selected_action[$item] = 1; } # Toggle on
  }

# Format the return string based on selections.
  $ret = "";
  $i = 1;  
  while ($i < $menu_index) {
    if ($selected_action[$i]) { $ret .= "$menu_sel_action[$i],"; }
    $i++;
  }
  chop($ret);       # Remove final comma
  @selected_action = ();
  if ($ret eq "") { $ret = "%NONE%"; }
  &menu_return_prep;
  $ret;
}

#**********
#  MENU_GETSTR
#
#  Function:	Prompt for a string (allowing editing) from one row of the
#		screen.
#
#  Call format:	$string = &menu_getstr(row,col,"Prompt text",clr_flag);
#
#  Arguments:	- Row,Col for prompt and data-input.
#		- Prompt text.
#		- Boolean flag (0=leave line on exit, 1=clear line on exit)
#
#  Returns:     String (can be null)
#
#  Notes:       1) This routine ALWAYS sets "nocbreak" and "echo" terminal
#                  modes before returning.
#		2) This routine checks for the right edge of the screen.
#**********
sub main'menu_getstr {
  local($row,$col,$prompt,$cleanup) = @_;
  local($prompt_col,$left_col,$action);
  local($string) = "";
  local($i) = 0;

# Set cbreak, noecho.
  &main'cbreak;
  &main'noecho;

# Add the prompt prompt
  &main'move($row,$col);
  &main'clrtoeol;
  &main'addstr($prompt);
  &main'refresh;

  $prompt_col = $col;
  $col = $col + length($prompt);
  $left_col = $col;

# Perform editing until "Return" pressed.
  while (1) {
    $action = &collect_seq;
    if (($action eq $kr) || ($action eq $ansi_kr)) {		# Right-arrow
      if ($i+1 > length($string)) { &bell; }
      else {
	$col++; $i++;
	&main'move($row,$col);
      }
    }
    elsif (($action eq $kl) || ($action eq $ansi_kl)) {		# Left-arrow
      if ($i-1 < 0) { &bell; }
      else {
	$col--; $i--;
	&main'move($row,$col);
      }
    }
    elsif (($action eq "\177") || ($action eq "\010")) {	# Delete/BS
      if ($i-1 < 0) { &bell; }
      else {
	$col--; $i--;
	&main'move($row,$col);
	&main'delch;
	$string = substr($string,0,$i).substr($string,$i+1);
      }
    }
    elsif (($action eq $cr) || ($action eq $nl) || 
	   ($action eq "\n")) {					# Terminate
      if ($cleanup) {
	&main'move($row,$prompt_col);		# Clear our stuff
	&main'clrtoeol;
	&main'refresh;
      }
      &main'nocbreak;		# ALWAYS turn off "cbreak" mode
      &main'echo;		# ALWAYS turn on "echo"
      return($string);
    }
    elsif (($action eq $ku) || ($action eq $ansi_ku) ||
	   ($action eq $kd) || ($action eq $ansi_kd)) {
      ; # Ignore
    } else {
      if ($col+1 > $main'COLS) { &bell; }
      else {
	&main'move($row,$col);		# Insert the character on the screen
	&main'insch(ord($action));
	$string = substr($string,0,$i).$action.substr($string,$i);
	$col++; $i++;
	&main'move($row,$col);
      }
    }
    &main'refresh;
  }
}

#**********
#  MENU_DISPLAY_INTERNAL 
#
#  Function:	Display items in menu_sel_text array, allow selection, and
#		return appropriate selection-string.
#
#  Call format:	$sel = &menu_display_internal([0|1|2],"Prompt text",
#			$arrow_pos,$top_item,"current");
#
#  Arguments:   - Menu type (0=simple, 1=radio, 2=multiple-select).
#               - Prompt text (for the bottom line of menu).
#		- Line number offset for the arrow (defaults to zero).
#		- Index of top item on screen (defaults to zero).
#		- Current selection (radio only).
#
#  Returns:     Selected action string (from second param on &menu_item)
#		%UP%    -- "u"|"U" pressed (or "t"|"T" and looking for top)
#               %EMPTY% -- Nothing in menu to display
#               %DONE%  -- Done with selections (radio and mult only)
#**********
sub menu_display_internal {
  ($menu_type,$menu_prompt,$arrow_spec,$menu_top_item,$radio_sel) = @_;
  local($i,$search,$low_sel);

# If looking for top menu, return with "%UP%".
  if ($finding_top) {
    if ($menu_is_top_one) { $finding_top = 0; } 
    else { return("%UP%"); }
  }

# Check for no "menu_item" calls.
  $total_items = $#menu_sel_text + 1;
  if ($total_items <= 0) {
    &main'nocbreak;		# ALWAYS turn off "cbreak" mode
    &main'echo;			# ALWAYS turn on "echo"
    return("%EMPTY%");
  }

  &main'cbreak;			# cbreak mode (each character available)
  &main'noecho;			# Menus are always "noecho"

  if ($total_items <= $items_per_screen) { $menu_single_page = 1; }
  else { $menu_single_page = 0; }

  if ($menu_prompt eq "") {
    $menu_prompt = "h)elp q)uit u)p";
    if ($menu_top_activated) {
      if (! $menu_is_top_one) { $menu_prompt .= " t)op"; }
    }
    if ($menu_type == 2) { $menu_prompt .= " a)ll m)atch c)lear"; }
    if (! $menu_single_page) {
      $menu_prompt .= " n)ext-pg p)rev-pg";
    }
    $menu_prompt .= " b)egin e)nd";
    if (length($menu_prompt)+9 < $main'COLS - 7) {
      $menu_prompt. = " r)efresh";
    }
  }
  if (length($menu_prompt) > $main'COLS - 7) {
    $menu_prompt = substr($menu_prompt,0,$main'COLS - 7);
  }

# Validate/adjust paramaters.
  $arrow_line = $arrow_spec + $first_line;
  if ($menu_top_item + 1 > $total_items) { $menu_top_item = $total_items - 1; }
  if ($arrow_line < $first_line) { $arrow_line = $first_line; }
  if ($arrow_line > $last_line) { $arrow_line = $last_line; }

# Clear screen and add top title and bottom prompt
  &menu_top_bot;
  $move_amt = 0;
  $number = 0;
   
  while (1) {
    $number_shown = $menu_top_item + $items_per_screen;
    if ($number_shown > $total_items) { $number_shown = $total_items; }
    $percent = int($number_shown * 100 /$total_items);

    &menu_page;			# Display current page
    &main'refresh;		# Update screen

# Collect key sequences until something we recoginize 
# (or we know we don't care)
    $action = &collect_seq;

#  Perform action based on keystroke(s) received
    $move_amt = 0;
    if ($action ne "") {
      $last_arrow_line = $arrow_line;
      if (($action eq $kd) || ($action eq $ansi_kd) ||		# Down-arrow
	  ($action eq $kr) || ($action eq $ansi_kr)) {		# Right-arrow
        $number = 0;
        if ($arrow_line < $max_sel_line) { $arrow_line++; }
        else {
	  if ($arrow_line == $last_line) { $move_amt = 1; }
        }
      }
      elsif (($action eq $ku) || ($action eq $ansi_ku) ||	# Up-arrow
	     ($action eq $kl) || ($action eq $ansi_kl)) {	# Left-arrow
        $number = 0;
        if ($arrow_line > $min_sel_line) { $arrow_line--; }
        else { $move_amt = -1; }
      }
      elsif (($action eq "n") || ($action eq "N") ||		# Next
	     ($action eq " ")) {
        $number = 0;
	$move_amt = $items_per_screen;
      }
      elsif (($action eq "b") || ($action eq "B")) {		# Begin
        $menu_top_item = 0;
	$arrow_line = $first_line;
      }
      elsif (($action eq "e") || ($action eq "E")) {		# End
	if (! $menu_single_page) {
	  $menu_top_item = $menu_index - $items_per_screen;
	}
	$arrow_line = $last_line;
      }
      elsif (($action eq "p") || ($action eq "P")) {		# Previous
        $number = 0;
	$move_amt = -$items_per_screen;
      }
      elsif (($action eq "a") || ($action eq "A")) {		# Select all
	if ($menu_type == 2) {
	  $i = 1;  
	  while ($i < $menu_index) { $selected_action[$i] = 1; $i++; }
	}
      }
      elsif (($action eq "c") || ($action eq "C")) {		# Clear all
	if ($menu_type == 2) {
	  $i = 1;  
	  while ($i < $menu_index) { $selected_action[$i] = 0; $i++; }
	}
      }
      elsif (($action eq "m") || ($action eq "M") ||		# Match string
             ($action eq "=") || ($action eq "/")) {
	if ($menu_type == 2) {
	  $search = &main'menu_getstr($last_line+1,0,"Search string: ",1);
	  &main'cbreak;			# menu_getstr turned this off
	  &main'noecho;			# menu_getstr turned this on
	  if ($search) {		# Toggle selections
	    $search =~ tr/A-Z/a-z/;
	    $i = 1;  
	    while ($i < $menu_index) {
	      $low_sel = $menu_sel_text[$i];
	      $low_sel =~ tr/A-Z/a-z/;
	      if (index($low_sel,$search) >=0) { $selected_action[$i] = 1; }
	      $i++;
	    }
	  }
	}
      }
      elsif (($action eq "r") || ($action eq "R")) {		# Refresh
	  &main'clear;
	  &menu_top_bot;
	  &menu_page;
	  &main'refresh;
      }
      elsif (($action eq "\177") || ($action eq "\010")) {	# Delete/BS num-reset
	  $number = 0;
	  $arrow_line = $first_line;
	  &menu_page;
	  &main'refresh;
      }
      elsif (($action eq "Q") || ($action eq "q")) {		# Quit
        &main'clear; $xrow = $xcol = 0;
	&main'move(0,0);
	&main'addstr("Do you really want to quit? y");
	&main'move(0,28);
	&main'refresh;
	$ch = &main'getch;
	if (($ch eq $cr) || ($ch eq $nl) || ($ch eq "\n")) { $ch = "y"; }
	$ch =~ tr/A-Z/a-z/;
	if ($ch eq "y") {
	  if ($menu_exit_routine ne "") { &$menu_exit_routine; }
	  &menu_return_prep;
	  exit(0);
	}
	&main'clear; $xrow = $xcol = 0;
	&menu_top_bot;		# Re-display current page
	&menu_page;
	&main'refresh;
      }
      elsif (($action eq "U") || ($action eq "u")) {		# Up
	$finding_top = 0;
	return("%UP%");
      }
      if (($action eq "T") || ($action eq "t")) {		# Top
        if (($menu_top_activated) && (! $menu_is_top_one)) {
	  $finding_top = 1;
	  return("%UP%");
	}
      }
      elsif (($action eq "h") || ($action eq "H") ||		# Help
	     ($action eq "?")) {
	&menu_show_help;	# Show help page
	&menu_top_bot;		# Clear and re-display the current page
	&menu_page;
	&main'refresh;
      }
      elsif ($action eq "!") {					# Shell escape
	if ($menu_shell_text ne "") {
	  &main'clear; $xrow = $xcol = 0;	# Clear screen
	  &main'refresh;
	  &main'nocbreak;	# ALWAYS turn off "cbreak" mode
	  &main'echo;		# ALWAYS turn on "echo"
	  $xrow = $xcol = 0;
	  &print_nl("Entering command shell via \"$menu_shell_text\".",1);
	  &print_nl("Return here via shell exit (normally Control-D).",1);
	  &main'refresh;
	  &main'endwin;
	  system($menu_shell_text);
	  &main'cbreak;		# cbreak mode (each character available)
	  &main'noecho;		# Menus are always "noecho"
	  &main'clear; $xrow = $xcol = 0;	# Clear screen
	  &menu_top_bot;	# Re-display the current page
	  &menu_page;
	  &main'refresh;
	}
      }
      elsif (($action eq $cr) || ($action eq $nl) || 
	     ($action eq "\n")) {				# RETURN
	if ($number) { $item = $number - 1; }
	else { $item = $menu_top_item + $arrow_line - $first_line; }
	if (($item < $menu_top_item) || ($item > $menu_bot_item)) {
	  &bell;
	  $number = 0;
	}
	else {
          if ($#_ > 0) { @_[2] = $arrow_line - $first_line; }
          if ($#_ > 0) { @_[3] = $menu_top_item; }
	  return(@menu_sel_action[$item]);
	}
      }
      else {
	$digit_val = index("0123456789",$action);		# Number
	if ($digit_val >= 0) {
	  $number = $number * 10 + $digit_val;
	  if ($number >= $menu_top_item + 1) { 
	    if ($number <= $menu_bot_item + 1) {
	      $arrow_line = $number - $menu_top_item + $first_line - 1;
	    } else {
	      &bell;
	      $number = 0;
	      $arrow_line = $first_line;
	    }
	    &menu_page;
	    &main'refresh;
	  }
	}
      }

# Check for paging of the menu text
      if ($move_amt != 0) {
	if ($move_amt < 0) { # Move backward
	  $menu_top_item = $menu_top_item + $move_amt;
	  if ($menu_top_item < 0) { $menu_top_item = 0; }
	}
	else { # Move forward
	  if ($menu_top_item + $move_amt < $total_items) {
	    $menu_top_item = $menu_top_item + $move_amt;
	  }
	} 
      }

# Erase the last selection arrow
      if ($menu_numbered) {
	&main'move($last_arrow_line,0);
	&main'addstr("  ");
      }
    }
  }
}

#**********
# COLLECT_SEQ -- Collect characters until a sequence we recognize (or we
#		 know it cannot possibly fit any "magic" sequences.
#**********
sub collect_seq {
  local($i,$possible);
  local($collect,$action) = "";

  $possible = $#magic_seq;	# Set number of possible matches 

seq_seek:
  while ($possible > 0) {
    $ch = &main'getch;
    $collect = $collect.$ch;
    $i = 0;
    $possible = 0;
    $action = $ch;
try:
    while ($i <= $#magic_seq) {
      if (length($collect) > length($magic_seq[$i])) {
	$i++;
	next try;
      }
      if (substr($magic_seq[$i],0,length($collect)) eq $collect) {
        $possible++;
	if ($collect eq $magic_seq[$i]) {
          $action = $magic_seq[$i];
          last seq_seek;
        }
      }
      $i++;
    } # end while
  }
  $action;
}

#**********
#  MENU_TOP_BOT -- Display top and bottom lines of current menu
#**********
sub menu_top_bot {
  &main'move(0,$menu_top_title_col);
  if ($menu_top_title_attr == 0) { &main'standout; }
  &main'addstr($menu_top_title);
  if ($menu_top_title_attr == 0) { &main'standend; }
  &main'move($last_line+2,7);
  &main'addstr($menu_prompt);
}

#**********
#  MENU_PAGE -- Display one page of menu selection items.
#**********
sub menu_page {
  local($flag) = 0;
  local($line);

# Update percentage on bottom line
  &main'move($last_line+2,0);
  &main'standout;
  if ($menu_single_page) { &main'addstr("(All) "); }
  else { &main'addstr(sprintf("\(%3d%%\)",$percent)); }
  &main'standend;
 
# Display current page of menu
  $item = $menu_top_item;
  $menu_bot_item = $menu_top_item;
  $curr_line = $first_line;
  $min_sel_line = $first_line;
  $max_sel_line = $first_line;
  while ($curr_line <= $last_line) {
    &main'move($curr_line,0);
    &main'clrtoeol;
    $line = "";
    if ($item < $total_items) {
      $sel_num = $item + 1;
      if ($menu_numbered) { $line .= &menu_add_number($sel_num); }
      if (($menu_type == 1) || ($menu_type == 2)) {
	if ($sel_num != 1) {
	  if ($menu_type == 1) {
	    if ($menu_sel_action[$item] eq $radio_sel) { $line .= "[X] "; }
	    else { $line .= "[ ] "; }
	  } else { # menu_type is 2
	    if ($selected_action[$item]) { $line .= "[X] "; }
	    else { $line .= "[ ] "; }
	  }
	}
      }
      $line .= $menu_sel_text[$item];
      if (length($line) > $main'COLS - 1) { # Truncate lines that would wrap
	$line = substr($line,0,$main'COLS - 1);
      }
      &main'addstr($line);
      $max_sel_line = $curr_line;
    }
    $item++;
    $curr_line++;
  }
  $menu_bot_item = $item - 1;
 
# Position the selection arrow on the screen (if numbered menu)
  if ($arrow_line > $max_sel_line) { $arrow_line = $max_sel_line; }
  &main'move($arrow_line,0);
  if ($menu_numbered) { &main'addstr("->"); }
}

#**********
# MENU_ADD_NUMBER -- Format selection number.
#**********
sub menu_add_number {
  local($sel_num) = @_;
  local($sel_str) = "  ";

  if ($sel_num < 1000) { $sel_str .= " "; }
  if ($sel_num < 100) { $sel_str .= " "; }
  if ($sel_num < 10) { $sel_str .= " "; }
  $sel_str .= "$sel_num) ";
  $sel_str;
}

#**********
# MENU_RETURN_PREP -- Common return functions.
#**********
sub menu_return_prep {
  &main'nocbreak;
  &main'echo; 
  &main'clear;  $xrow = $xcol = 0;
  &main'refresh; 
  if (!$curses_application) { &main'endwin; }
}

#**********
# MENU_SHOW_HELP
#*********
sub menu_show_help {
  &main'clear; $xrow = $xcol = 0;
  &main'refresh; 
  &print_nl("--------------------------------",1);
  &print_nl("Menu Help (perlmenu version 2.1)",1);
  &print_nl("--------------------------------",2);
  if ($menu_type == 1) {
    &print_nl("- Use up/down arrows to place \"->\" in front of your selection.",1);
    &print_nl("- Press Return (or Enter) to chose that selection.",1);
    &print_nl("- Select the first item when ready to continue.",2);
  }
  elsif ($menu_type == 2) {
    &print_nl("- Use up/down arrows to place \"->\" in front of your selection.",1);
    &print_nl("- Press Return (or Enter) to toggle the selection on/off.",1);
    &print_nl("- Select the first item when ready to continue.",2);
  } else {
    &print_nl("- Use up/down arrows to place \"->\" in front of your selection.",1);
    &print_nl("- Press Return (or Enter) when ready to continue.",2);
  }
  &print_nl("Available action-keys:",1);
  &print_nl("h - Show this help screen.",1);
  &print_nl("q - Quit entirely.",1);
  &print_nl("u - Return to the previous menu or function.",1);
  if (($menu_top_activated) && (! $menu_is_top_one)) {
    &print_nl("t - Return to the top menu.",1);
  }
  if ($menu_type == 2) {
    &print_nl("a - Select all items.",1);
    &print_nl("m - Select based on a case-insensitive string match.",1);
    &print_nl("c - Clear all selections.",1);
  }
  if (! $menu_single_page) {
    &print_nl("n - Move forward to next page.",1);
    &print_nl("p - Move backward previous page.",1);
  }
  &print_nl("b - Move to the item at the beginning of the menu.",1);
  &print_nl("e - Move to the item at the end of the menu.",1);
  &print_nl("r - Refresh the screen.",1);
  if ($menu_shell_text ne "") {
    &print_nl("! = Enter command shell via \"$menu_shell_text\".",1);
  }
  &print_nl(" ",1);
  &main'addstr("[Press any key to continue]");
  &main'refresh;
  $ch = &main'getch;
  &main'clear; $xrow = $xcol = 0;
  &main'refresh;
}

sub print_nl {
  local($text,$skip) = @_;

  &main'addstr($text);
  if ($skip) { &nl($skip); }
  &main'refresh;
}

sub nl {
  local($skip) = @_;
  $xrow += $skip;
  $xcol = 0;
  if ($xrow > $main'LINES - 1) {
    &main'refresh;
    &main'clear; $xrow = 0; $xcol = 0;
    &main'refresh;
  }
  &main'move($xrow,$xcol);
}

sub defbell {
  eval q#
    sub bell { print "\007"; }
  #;
}

1;
